! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

!
!   caltest_epn.f90  --- f90 program to test the CAL (EPN aspects only)
!
!   DESCRIPTION:
!       This program tests the correctness of the EPN-related CAL part by
!       performing a sequence of subroutine calls followed by
!       checks of the output. If any of these checks fails, the
!       program shall exit with a non-zero return value. If all tests
!       succeed 0 will be returned to the calling environment.
!
program caltest_epn

use cal
use errorhandling

implicit none
    logical                        :: isClose
    integer                        :: i, j, nx, ny
    integer(int16), dimension(:), pointer :: &
                                   rawX => null(), rawY => null(), &
                                   badPixType => null(), yextent => null()
    logical, dimension(:), pointer :: uplinked => null()
    integer(int16), dimension(0:5) :: &
                                   rawXtest = (/1, 23, 34, 56, 64, 33/),&
                                   rawYtest = (/1, 200, 111, 139, 32, 200/),&
                                   Ein = (/123, 456, 789, 999, 364, 888/), &
                                   pattId = (/0, 1, 2, 4, 1, 0/), &
                                   rawXtest2, rawYtest2
    integer(int32), dimension(0:5) :: frame = (/1, 1, 1, 1, 1, 1/)

    real(single), dimension(:), pointer :: EinReal => null(), & 
                                           EoutReal => null(), EoutReal2 => null()
    integer(int16), dimension(:), pointer :: EoutInt => null(), EoutInt2 => null()
    real(double), dimension(:), pointer :: xmilli => null(), ymilli => null(), &
                                           zmilli => null()
    integer(int16), dimension(:,:), pointer :: ibadpixmap => null()

    type(BoresightT)                       :: boresight
    real(double), dimension(:,:), pointer  :: matrix => null()
    real(single), dimension(:,:), pointer  :: PSFmap => null()
    real(single), dimension(:), pointer :: parOut => null()
    type(PsfT)                      :: psf
    type(PatternFractionServerT)    :: pfds;
    type(CurveT)                    :: curve
    type(PsfWindowT)                :: window
    real(double), dimension(:), pointer :: pattFracChan => null()
    real(double)                    :: focalLength, plateScaleX, dum
    character(len=*), parameter     :: myBadPixFile = &
                                      './epn_badpix.ccf'
    type(CalInfo)                   :: badPixFileInfo
    integer(int32)                  :: instrumentOut, ccdChipIdOut, &
                                       accLevelOut, modeNr
    integer(int32)                  :: ndsclin
    integer(int16)                  :: mipsel, maxmip
    real(single)                    :: sum
    real(double)                    :: rdp

    call CAL_setState('epn.fits')
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      accuracyLevel = accLevelOut)
    if (instrumentOut /= EPN) &
        call fatal('caltest_epn', "instrumentOut /= EPN")
    if (ccdChipIdOut /= EPN_CCD_12) &
        call fatal('caltest_epn', "ccdChipIdOut /= EPN_CCD_12")

    call CAL_setState(EPN, ccdChipId=EPN_CCD_1, &
                      accuracyLevel = ACCURACY_LOW)
    call CAL_getState(instrument=instrumentOut, ccdChipId = ccdChipIdOut, &
                      accuracyLevel = accLevelOut)
    call CAL_setState(temperatureCamera=299.d0)
    if (instrumentOut /= EPN) &
        call fatal('caltest_epn', "instrumentOut /= EPN")
    if (ccdChipIdOut /= EPN_CCD_1) &
        call fatal('caltest_epn', "ccdChipIdOut /= EPN_CCD_1")
    if (accLevelOut /= ACCURACY_LOW) &
        call fatal('caltest_epn', "accLevelOut /= ACCURACY_LOW")

    !
    !   gain correction (low+high accuracy)
    !
    call CAL_setState(ccdChipId = EPN_CCD_1, randomize = .false., &
                      accuracyLevel = ACCURACY_LOW)

    call CAL_integerToReal(Ein, EinReal)
    call CAL_pnGainCorrect(rawXtest, rawYtest, EinReal, EoutReal)
    call CAL_realToInteger(EoutReal, EoutInt)
    write (*, *) (EoutInt(i), i=0,size(EoutInt)-1)
    if (EoutInt(0) /= 119 .or. &
        EoutInt(1) /= 445 .or. &
        EoutInt(2) /= 775 .or. &
        EoutInt(3) /= 1000 .or. &
        EoutInt(4) /= 367 .or. &
        EoutInt(5) /= 863) &
        call fatal('caltest_epn', "low-acc. gain correction does not work")
    call CAL_releaseMemory(EinReal)
    call CAL_releaseMemory(EoutInt)

    !
    !   CTI correction
    !
    rawXtest2(0) = EPN_MIN_X
    rawXtest2(1) = EPN_MAX_X
    rawXtest2(2) = EPN_MAX_X
    rawXtest2(3) = EPN_MIN_X
    rawXtest2(4) = EPN_MIN_X
    rawXtest2(5) = EPN_MIN_X
    rawYtest2(0) = EPN_MIN_Y
    rawYtest2(1) = EPN_MIN_Y
    rawYtest2(2) = EPN_MAX_Y
    rawYtest2(3) = EPN_MAX_Y
    rawYtest2(4) = EPN_MIN_Y
    rawYtest2(5) = EPN_MAX_Y

    call CAL_setState(temperatureTracking = .true.)
    call CAL_pnCtiCorrect(rawXtest2, rawYtest2, frame, EoutReal, EoutReal2)
    call CAL_realToInteger(EoutReal2, EoutInt2)
    write (*, *) (EoutInt2(i),i=0,size(EoutInt2)-1)

    ! long-term gain
    write(*,*) 'Long term gain'
    call CAL_setState(ccdModeId = PRIME_FULL_WINDOW)
    EoutReal2(0) = 3.0 / 5E-3
    EoutReal2(1) = 6.0 / 5E-3

    write(*,*) 'In: ',EoutReal2(0),EoutReal2(1)
    call CAL_pnAdditionalGainCtiCorrect(EoutReal2,rawYtest2,&
                               165._double,parOut, &
                               .true.,.false.,.false.,.true.,.true.)
    write(*,*) 'Out: ',EoutReal2(0),EoutReal2(1)

    ! Background correction
    ndsclin=100;
    mipsel=1;
    maxmip=5;
    call CAL_pnBkgDepGainCorrect(EoutReal2, ndsclin, mipsel, maxmip);
    write(*,*) 'OutBKG5: ',EoutReal2(0),EoutReal2(1),mipsel, maxmip
    maxmip=63
    call CAL_pnBkgDepGainCorrect(EoutReal2, ndsclin, mipsel, maxmip);
    write(*,*) 'OutBKG63: ',EoutReal2(0),EoutReal2(1),mipsel, maxmip

    call CAL_pnGainCorrect(rawXtest2, rawYtest2, EoutReal, EoutReal2)
    EoutReal2(0) = 0.01 / 5E-3 / 0.9665
    EoutReal2(1) = 0.8 / 5E-3 / 0.9665
    EoutReal2(2) = 1.5 / 5E-3 / 0.9665
    EoutReal2(3) = 3.7 / 5E-3 / 0.9665
    EoutReal2(4) = 5.0 / 5E-3 / 0.9665
    EoutReal2(5) = 6.0 / 5E-3 / 0.9665
    write(*,*) 'In: ',EoutReal2(0),EoutReal2(1)
    call CAL_pnAdditionalGainCtiCorrect(EoutReal2,rawYtest2,&
                               165._double,parOut, &
                               .true.,.false.,.true.,.true.,.true.)
    write(*,*) 'Out: ',EoutReal2(0),EoutReal2(1)
    write(*,*) 'Long term gain'

    write(*,*)' Testing Spatial CTI correction'
    call CAL_pnspatialCtiCorrect(EoutReal2,rawXtest2,rawYtest2)
    write(*,*)' FInished Spatial CTI correction'
    !
    write(*,*)' Testing combined event CTI correction - energies in eV'
    EoutReal2(0) = 0.01 / 1E-3 
    EoutReal2(1) = 0.8 / 1E-3
    EoutReal2(2) = 1.5 / 1E-3
    EoutReal2(3) = 3.7 / 1E-3
    EoutReal2(4) = 5.0 / 1E-3
    EoutReal2(5) = 6.0 / 1E-3
    write(*,*)'In: ',(EoutReal2(i),i=0,5)
    write(*,*)'In: ',(pattId(i),i=0,5)
    call CAL_pnCombinedEventCorrect(EoutReal2, pattId)
    write(*,*)' FInished combined event CTI correction'
    write(*,*)'Out: ',(EoutReal2(i),i=0,5)

    !
    call CAL_releaseMemory(EoutInt2)
    ! call CAL_releaseMemory(EoutReal2)

    call CAL_setState(ccdModeId = PRIME_FULL_WINDOW)

    call CAL_setState(accuracyLevel = ACCURACY_LOW)

    !
    !   re-emission threshold
    !
    call CAL_setState(temperature = 165._double)
    write (*, *) CAL_pnReEmissionThreshold(1228.965_double, 150_int16);
    if (.not. isClose(real(CAL_pnReEmissionThreshold(1228.965_double, &
                          150_int16)), 26.499031842)) &
        call fatal('caltest_epn', "Re-emission threhold calc does not work")

    !
end program


logical function isClose(value, trueval)
use types

    real(single), intent(in)    :: value, trueval
    real(single)                :: absval, abstrueval

    isClose = .true.
    if (value * trueval .lt. 0.) isClose = .false.
    absval = abs(value)
    abstrueval = abs(trueval)
    if (absval .lt. .99999 * abstrueval .or. &
        absval .gt. 1.0001 * abstrueval) isClose = .false.
end function
