// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	BackgroundDataServer.cc --- implementation of BackgroundDataServer
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2002-06-14: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers, Fri Jun 14 16:31:10 MEST 2002
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/BackgroundDataServer.cc,v 1.3 2009/07/06 14:25:19 aibarra Exp $
//

#include "BackgroundDataServer.h"

using namespace std;

//
//	----------------------( Background )-----------------------------------
//
/* AI gcc4 porting
 * too few template-parameter-lists => add template<>
 */
template<>
CalAtomCollection<BackgroundDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;

	_nCCDs	= instrument->nCcds();
	_nNodes	= instrument->nNodes();
	_nModes	= instrument->nModes();
	assert(_nCCDs && _nNodes && _nModes);
	_nEntries	= 1;
	BackgroundDataServer	*null = 0;
	_collection = new vector<BackgroundDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::Background);
	DataSet *dataSet = e->open();

	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
		case Xmm::EPN: {
			CcfConstituentTable ccftabParticle(dataSet,
										e->queryBlockName("PARTICLE")),
								ccftabDiffuse(dataSet,
										e->queryBlockName("X_RAY_DIFFUSE")),
								ccftabDiffuseMod(dataSet,
								e->queryBlockName("X_RAY_DIFFUSE_MODULATION"));

			(*_collection)[0] = new EpicBackgroundDataServer(ccftabParticle,
															 ccftabDiffuse,
															 ccftabDiffuseMod);
			break;
		}
		case Xmm::RGS1:
		case Xmm::RGS2:
			(*_collection)[0] = new RgsBackgroundDataServer();
			break;
		default:
			errstr<<"No "<<BackgroundDataServer::id()<<" for instrument "
				  <<instrument->id()
				  <<Sas::error("NoSuchCcf");
			break;
	}

	dataSetServer->close(dataSet);
} // CalAtomCollection<BackgroundDataServer>::CalAtomCollection

/* AI gcc4 porting
 * too few template-parameter-lists => add template<>
 */
template<>
BackgroundDataServer *CalAtomCollection<BackgroundDataServer>::select(
														Ccd, Node, unsigned)
const
{
	return (*_collection)[0];
} // CalAtomCollection<BackgroundDataServer>::select()


BackgroundSpectrum BackgroundDataServer::particle(double, double) const
{
	errstr<<"Generic BackgroundDataServer cannot serve particle spectrum"
		  <<Sas::error("UnspecializedObject");

	return BackgroundSpectrum();
}

BackgroundSpectrum BackgroundDataServer::diffuseX(double, double) const
{
	errstr<<"Generic BackgroundDataServer cannot serve diffuse X-ray spectrum"
		  <<Sas::error("UnspecializedObject");

	return BackgroundSpectrum();
}

HtrBackgroundIntensity BackgroundDataServer::betaSelect(double beta) const
{
	errstr<<"Generic BackgroundDataServer cannot serve HtrBackgroundIntensity"
		  <<Sas::error("UnspecializedObject");

	return HtrBackgroundIntensity(beta);
}


//
//	instrument specific
//
EpicBackgroundDataServer::EpicBackgroundDataServer(
										const CcfConstituentTable &particle,
										const CcfConstituentTable &diffuse,
										const CcfConstituentTable &diffuseMod,
							 CalAtomCollection<BackgroundDataServer> *parent)
	: CalChild<CalAtomCollection<BackgroundDataServer> >(parent),
	  BackgroundDataServer(parent)
{
	//
	//	read particle background parameters
	//
	const Table *t1 = particle.block();
	ColumnData *cdpar = t1->column("PAR")->data();
	real32	*par = cdpar->real32Data();

	_particleP0 = par[0];
	_particleP1 = par[1];

	//
	//	read diffuse X-ray components
	//
	const Table *t2 = diffuse.block();
	ColumnData *d = t2->column("SPECTRUM")->data();
	unsigned nChanns = d->elements();

	for (unsigned n=0; n<nChanns; ++n) {
		unsigned long pos = d->position(0, n);
		_diffuse.pushChannel(d->real32Data()[pos]);
	}

	//
	//	read diffuse X-ray modulation part
	//
	const Table *t3 = diffuseMod.block();
	ColumnData *cdt3 = t3->column("PAR")->data();
	par = cdt3->real32Data();
	_diffModP0 = par[0];
	_diffModP1 = par[1];
	_diffModP2 = par[2];

	//	tidy up
	delete cdt3;
	delete d;
	delete cdpar;
} // EpicBackgroundDataServer::EpicBackgroundDataServer()


BackgroundSpectrum EpicBackgroundDataServer::particle(double time, double r)
const
{
	errstr<<"Background::particle: dependence on time ("<<time<<")"
		  <<" and r ("<<r<<") not yet implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	BackgroundSpectrum particleBackground;
	const vector<unsigned> &diffChanns = _diffuse.channels();
	unsigned nChanns = _diffuse.size();
	for (unsigned n=0; n<nChanns; ++n)
		particleBackground.pushChannel(_particleP0 +
									   _particleP1 * DBL(diffChanns[n]));

	return particleBackground;
} // EpicBackgroundDataServer::::particle()


BackgroundSpectrum EpicBackgroundDataServer::diffuseX(double time, double r)
const
{
	errstr<<"Background::particle: dependence on time ("<<time<<")"
		  <<" and r ("<<r<<") not yet implemented"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	BackgroundSpectrum diffuseBackground(_diffuse);
	//
	//	modify
	//
	double mod = _diffModP0 + _diffModP1 * r + _diffModP2 * r * r;
	unsigned nChanns = diffuseBackground.size();
	for (unsigned n=0; n<nChanns; ++n)
		diffuseBackground[n] *= mod;

	return diffuseBackground;
} // Background::diffuseX


EpicBackgroundDataServer::~EpicBackgroundDataServer() {}

double HtrBackgroundIntensity::accumulatedIntensity(double PIfrom,
													double PIto) const
{
	return PIto - PIfrom;	//	for the moment
} //  HtrBackgroundIntensity::accumulatedIntensity()

RgsBackgroundDataServer::RgsBackgroundDataServer(
							 CalAtomCollection<BackgroundDataServer> *parent)
	: CalChild<CalAtomCollection<BackgroundDataServer> >(parent),
	  BackgroundDataServer(parent) {}

HtrBackgroundIntensity RgsBackgroundDataServer::betaSelect(double beta) const
{
	return HtrBackgroundIntensity(beta);
} // RgsBackgroundDataServer::betaSelect()

RgsBackgroundDataServer::~RgsBackgroundDataServer() {}

