// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2020 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	FlareImageDataServer.cc --- implementation of class to return data from
//			     the Filter Wheel Closed data CCF
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 22-11-19: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2019-11-19
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//

#include "FlareImageDataServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

typedef multimap<const pair<string,string>,SImage>::iterator mapIter;

//
//	----------------------( mode parameters )-----------------------------
//

//  Define the FlareImageDataServer class
//
FlareImageDataServer::FlareImageDataServer( 
        multimap<const pair<string,string>,SImage> flareImage,
        CalAtomCollection<FlareImageDataServer> *parent)
        : CalChild<CalAtomCollection<FlareImageDataServer> >(parent), _flareImage(flareImage)
{
}
// Return the flare image for this combination
const SImage FlareImageDataServer::getFlareImage(string &band)
{
  SImage fim;

  string flarefilt = 
       Xmm::toString(calServer.state()->filter()->get());
  //cout<<"Filt: "<<flarefilt<<endl;

  // Extract the band,filter pair
  pair<string,string> p = make_pair(band, flarefilt);

  // Return the image for this combination
  mapIter s_it;
  pair<mapIter, mapIter> keyRange = _flareImage.equal_range(p);

  // Test for match
  if (keyRange.first == _flareImage.end() && 
      keyRange.second == _flareImage.end()) 
     errstr<<"The CCF does not contain Flare data "
      "for the band, filter combination "<<band<<
      " "<<flarefilt<<Sas::error("NoBandFilterData");
      
  for (s_it = keyRange.first;  s_it != keyRange.second;  ++s_it)
        {
           fim = (*s_it).second;
        }
  
  return fim;
}

// Return the flare image
//SImage *FlareImageDataServer::mkflareImage(unsigned dx, 
//                           unsigned dy, vector<float> v) 
//{
//        SImage *ffp = new SImage(dx, dy);
//        SImage &ff = *ffp;

//        for (unsigned nx=0; nx<dx; ++nx)
//                for (unsigned ny=0; ny<dy; ++ny)
//                        ff[nx][ny] = v[dx * (nx-1) +ny];

//        return ffp;
//} // OmFlatFieldServer::largeScaleFlatFieldMap()
//
// Define the entry point
// 
template<>
CalAtomCollection<FlareImageDataServer>::CalAtomCollection(
	                                     const XmmInstrument *instrument,
					     const Ccf *ccf,
					     Xmm::Filter filter,
					     AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
        _nCCDs		= instrument->nCcds();

	FlareImageDataServer *null = 0;
	_collection = new vector<FlareImageDataServer *>(1, null);

	CifEntry *e = ccf->entry(instrument, Ccf::FLARE);
	DataSet *dataSet = e->open();

      // Open the flare image extension
        Table *table = dataSet->table(e->queryBlockName("FLAREMAP"));

      // Read band, filter and energy into vectors
        vector<string> band, fimgfilt;
        vector<float> energy;
        //vector<vector<float> > vflareimg;
        //vector<unsigned long> extents = ad->parent()->size();
        //TImasterOffClosed = new SImage(extents[0], extents[1], ad->real32Data());
        ColumnData *insCol = table->column("INSTRUME")->data();
        CalOalUtility::fillVector(table, "BAND", band);
        CalOalUtility::fillVector(table, "FILTER", fimgfilt);
        CalOalUtility::fillVector(table, "ENERGY", energy);
        
        // Read the flare image
        ColumnData *cd = table->column("FLAREMAP")->data();
        real32* fimgd = cd->real32Data();

        // Get current instrument as a string
        string instrumentName(Xmm::toString(instrument->id()));

        //fimgd[608400*3+780] = 3.0;
        //cout<<"Pix: "<<fimgd[608400*3+780]<<endl;
        
        //CalOalUtility::fillVector(table, "FLAREIMG", vflareimg);
        //vector<unsigned> extents = cd->size();
        //unsigned nx=extents[0], ny=extents[0];
        //FIND NX, NY SOMEHOW
        unsigned nx=780, ny=780; // Bit of a hack this

      // Produce a map with the key defined by the band,filter pair
        multimap<const pair<string,string>, SImage> m;
   
      // Loop over each record
        for (unsigned i=0; i<table->rows(); ++i) {      

           string instId(uppercase(stringColumnRow(insCol, i)));
           if (instId == instrumentName) {

             // Make the key as a pair of band,filter
             pair<string,string> p = make_pair(band[i],fimgfilt[i]);
             //cout<<"Set: "<<band[i]<<" "<<fimgfilt[i]<<endl;
             SImage fim(nx,ny);

             int offset=i*nx*ny; // First pixel of the next map

             // Fill the image from the vector
             for (unsigned ix=0; ix<nx; ++ix)
               for (unsigned iy=0; iy<ny; ++iy)
                   fim[ix][iy] = fimgd[offset + nx*ix +iy];

             m.insert(make_pair(p, fim));
           }
        }

      // Create the collection - just one
        (*_collection)[0] = new FlareImageDataServer(m, this);

      // Delete the column pointers
      /*
        delete cdmodeId;
      */
	dataSetServer->close(dataSet);
} // CalAtomCollection<FlareImageDataServer>::CalAtomCollection


/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
FlareImageDataServer *CalAtomCollection<FlareImageDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
        //unsigned n = index(modeOffset);

        if (!(*_collection)[0])
           errstr<<"The CCF does not contain Flare data "
                          <<Sas::error("NoData");

	return (*_collection)[0];
} // CalAtomCollection<FlareImageDataServer>::select
