// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2020 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	SPDetImageDataServer.cc --- implementation of class to return data from
//			     the Filter Wheel Closed data CCF
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		RDS, 22-11-19: 1.0.0: - original
//
//	AUTHOR:
//		R. Saxton, 2019-11-19
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//

#include "SPDetImageDataServer.h"
#include "CalOalUtility.h"
#include "CalServer.h"

#include <algorithm>
#include <cmath>

using namespace CalOalUtility;
using namespace std;

//
//	----------------------( mode parameters )-----------------------------
//

//  Define the SPDetImageDataServer class
//
SPDetImageDataServer::SPDetImageDataServer( 
        SImage spdetImage,
        CalAtomCollection<SPDetImageDataServer> *parent)
        : CalChild<CalAtomCollection<SPDetImageDataServer> >(parent), _spdetImage(spdetImage)
{
}
// Return the flare image for this combination
const SImage SPDetImageDataServer::getSPDetImage()
{
  return _spdetImage;
}

template<>
CalAtomCollection<SPDetImageDataServer>::CalAtomCollection(
	                                     const XmmInstrument *instrument,
					     const Ccf *ccf,
					     Xmm::Filter filter,
					     AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nModes		= instrument->nModes();
        _nCCDs		= instrument->nCcds();

	SPDetImageDataServer *null = 0;
	_collection = new vector<SPDetImageDataServer *>(1, null);

	CifEntry *e = ccf->entry(instrument, Ccf::SPDETMAP);
	DataSet *dataSet = e->open();

      // Open the flare image extension
        Table *table = dataSet->table(e->queryBlockName("SPDETMAP"));

        ColumnData *insCol = table->column("INSTRUME")->data();
        
        // Read the swcx image
        ColumnData *cd = table->column("SPDETMAP")->data();
        real32* fimgd = cd->real32Data();


        // Get current instrument as a string
        string instrumentName(Xmm::toString(instrument->id()));

        //FIND NX, NY SOMEHOW
        unsigned nx=780, ny=780; // Bit of a hack this

        // Create the output image to store
        SImage fim(nx,ny);

        for (unsigned n=0; n<table->rows(); ++n) {
            string instId(uppercase(stringColumnRow(insCol, n)));
            if (instId == instrumentName) {


               int offset=n*nx*ny; // First pixel of the next map

               // Fill the image from the vector
               for (unsigned ix=0; ix<nx; ++ix)
                   for (unsigned iy=0; iy<ny; ++iy)
                      //fim[ix][iy] = fimgd[offset + nx*ix +iy];
                      fim[ix][iy] = fimgd[offset + ny*iy +ix];
            }
        }

      // Create the collection - just one
        (*_collection)[0] = new SPDetImageDataServer(fim, this);

      // Delete the column pointers
      /*
        delete cdmodeId;
      */
	dataSetServer->close(dataSet);
} // CalAtomCollection<SPDetImageDataServer>::CalAtomCollection


/* AI gcc4 porting
 *  too few template-parameter-lists => template<>
 */
template<>
SPDetImageDataServer *CalAtomCollection<SPDetImageDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
        //unsigned n = index(modeOffset);

        if (!(*_collection)[0])
           errstr<<"The CCF does not contain SPDet data "
                          <<Sas::error("NoData");

	return (*_collection)[0];
} // CalAtomCollection<SPDetImageDataServer>::select
