// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrector --- A class TimeCorrector is defined
//
//	DESCRIPTION:
//		Instances of class TimeCorrector can make fine time adjustments
//		of frame readout times.
//
//	CHANGE HISTORY:
//		UL, 1999-11-15: 1.0.0: - first version
//		UL, 2000-02-08: 1.1.0: - fixed compilation problem with gcc-2.95
//
//	AUTHOR:
//		U. Lammers, Mon Nov 15 15:20:13 MET 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Time resolution capabilities of the XMM EPIC pn-CCD in different
//			 readout modes, M. Kuster et al.
//
//	$Header: /sasbuild/sas_git_repo/cvs_copy/sasdev/sas/packages/cal/src/TimeCorrector.h,v 1.10 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef TIME_CORRECTOR_H
#define TIME_CORRECTOR_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalDataVector.h"

#include <vector>

//
//	---( TimeCorrector )--------------------------------------------------
//
class TimeCorrector : public CalAtom,
			  public virtual CalChild<CalAtomCollection<TimeCorrector> > {
public:
	TimeCorrector(CalAtomCollection<TimeCorrector> *parent = 0)
		: CalChild<CalAtomCollection<TimeCorrector> >(parent) {}
		//
		//	PURPOSE:
		//		Construct TimeCorrector object
		//


	static unsigned id() { return Ccf::TimeCorr; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(std::ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual std::vector<double> frameTimes() const;
	virtual double frameTime() const;
		//
		//	PURPOSE:
		//		Returns CCD readout frame times [ms] for all applicable
		//		modes or just the active one (second variant).
		//		The value is applicable to active CCD set in CAL state
		//


	virtual CalReal64Vector &correct(CalReal64Vector &time) const;
	virtual CalReal64Vector &correct(CalReal64Vector &time,
									 CalInt16Vector &rawy) const;
	virtual CalReal64Vector &correct(CalReal64Vector &time,
									 CalInt16Vector &rawy,
									 int16 sourceY) const;
		//
		//	PURPOSE:
		//		Assuming an event time has been assigned the telemetry time
		//		word as arrival time these routines do a mode-dependent
		//		adjustment of that time to give the best possible estimate
		//		of the actual photon arrival time (see description of
		//		instrument specific routines for details).
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			rawy	: Y position of the event (in PIXCOORD1 frame)
		//			sourceY	: Y position of the source (in PIXCOORD1 frame)
		//
		//	RETURNS:
		//		vector with corrected time tags
		//


	virtual const std::vector<std::vector<double> > &
												correctionParameters() const;
		//	implemented in EPN


	virtual ~TimeCorrector() {}
		//
		//	PURPOSE:
		//		Destruct TimeCorrector object
		//

private:
	void noOp() const;
};


//
//	a version of the TimeCorrector specialized for EPN
//
class EpnTimeCorrector : public virtual TimeCorrector {
public:
	EpnTimeCorrector(const CcfConstituentTable &frameTimeTab,
					 const CcfConstituentTable &timeCorrTab,
					 CalAtomCollection<TimeCorrector> *parent = 0);

	std::vector<double> frameTimes() const { return _frameTimes; }

	CalReal64Vector &correct(CalReal64Vector &time) const;
	CalReal64Vector &correct(CalReal64Vector &time,
							 CalInt16Vector &rawy) const;
	CalReal64Vector &correct(CalReal64Vector &time,
							 CalInt16Vector &rawy,
							 int16 sourceY) const;
		//
		//	PURPOSE:
		//		Fine-time correct event arrival times according to the
		//		recipes in [2]. The first form is applicable to all
		//		imaging modes (EpnPrimeFullWindow, PrimeFullWindowExtended,
		//		PrimeLargeWindow, PrimeSmallWindow) and FastBurst mode
		//		when the location of the source PSF on the CCD is unknown
		//		([1], Eq. 10). The second form is applicable to
		//		FastTiming mode data in the case that the PSF position
		//		is unknown ([1], Eq. 7). The third form applies to
		//		FastTiming/FastBurst mode data only. If the PSF position
		//		is known this improves the time resolution according to
		//		Eq. 9/11 in [1].
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			rawy	: Y position of the event (in PIXCOORD1 frame)
		//			sourceY	: Y position of the source (in PIXCOORD1 frame)
		//
		//	RETURNS:
		//		vector with corrected time tags
		//	

	const std::vector<std::vector<double> > &correctionParameters() const
	{ return _coefs; }

	~EpnTimeCorrector() {}

private:
	std::vector<std::vector<double> >	_coefs;
	std::vector<double>					_frameTimes;
};


//
//	a version of the TimeCorrector specialized for EMOS
//
class EmosTimeCorrector : public virtual TimeCorrector {
public:
	EmosTimeCorrector(const CcfConstituentTable &frameTimeTab,
					  CalAtomCollection<TimeCorrector> *parent = 0);

	EmosTimeCorrector(CalAtomCollection<TimeCorrector> *parent)
		: CalChild<CalAtomCollection<TimeCorrector> >(parent),
		  TimeCorrector(parent) {}

	std::vector<double> frameTimes() const;

	~EmosTimeCorrector() {}

private:
	std::vector<std::vector<double> > _frameTimes;
};


//
//	a version of the TimeCorrector specialized for RGS
//
class RgsTimeCorrector : public virtual TimeCorrector {
public:
	RgsTimeCorrector(CalAtomCollection<TimeCorrector> *parent)
		: CalChild<CalAtomCollection<TimeCorrector> >(parent),
		  TimeCorrector(parent) {}

	~RgsTimeCorrector() {}
};

#endif
